/* o_ovl.c - Overlap optimizations - Don Yang / Dave Newquist / John Rapp

   07/28/01
                                                                           */

#include<assert.h>
#include<stdio.h>
#include"token.h"
#include"tree.h"


/* Local functions */
static void BitToTag(Token *rtoken, int attribute);
static int GetNestedAttributes(Node *node);
static int PrepNeighbors(Node **node);
static void RotateAttr(Node **node, Token *tag);
static int TagToBit(Token *token);

/*************************************************************** o_overlap */
int o_overlap(Node **node)
{
   int modified;

   modified = 0;
   while( *node )
   {
      if( (*node)->tree )
      {
         modified |= PrepNeighbors(node);
         if( (*node)->tree )
            modified |= o_overlap(&((*node)->tree));
      }

      assert(*node != NULL);
      node = &((*node)->next);
   }
   return modified;

} /* o_overlap() */


/**************************************************************** BitToTag */
static void BitToTag(Token *rtoken, int attribute)
{
   static TokenType t[24] = {
      TagSize, TagSize, TagSize, TagSize, TagSize,
      TagSize, TagSize, TagSize, TagSize, TagSize,
      TagB, TagI, TagS, TagU, TagEM, TagTT,
      TagColor, TagColor, TagColor, TagColor,
      TagColor, TagColor, TagColor, TagColor};
   static int i[24] = {
      '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
      0, 0, 0, 0, 0, 0, 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w'};
   int b;

   for(b = 0; b < 24; b++)
   {
      if( attribute & (1 << b) )
      {
         rtoken->type = t[b];
         rtoken->lexeme.i = i[b];
         return;
      }
   }
   assert(0);

} /* BitToTag() */

/***************************************************** GetNestedAttributes */
static int GetNestedAttributes(Node *node)
{
   int attribute;

   attribute = TagToBit(node->token);
   for(node = node->tree; node != NULL; node = node->tree)
   {
      if( node->token->tagtype == 0 || node->next )
         break;
      attribute |= TagToBit(node->token);
   }

   return attribute;

} /* GetNestedAttributes() */

/*********************************************************** PrepNeighbors */
static int PrepNeighbors(Node **node)
{
   Token tag;
   int a;

   if( (*node)->next == NULL )
      return 0;

   /* Find common attributes between adjacent tags */
   a = GetNestedAttributes(*node) & GetNestedAttributes((*node)->next);
   if( a == 0 )
      return 0;
   BitToTag(&tag, a);

   /* Make common attributes the outer most tag */
   RotateAttr(node, &tag);
   RotateAttr(&((*node)->next), &tag);

   return 1;

} /* PrepNeighbors() */

/************************************************************** RotateAttr */
static void RotateAttr(Node **node, Token *tag)
{
   Node *r, *p, *t;

   p = NULL;
   for(r = *node; r; r = r->tree)
   {
      if( r->token->type == tag->type && r->token->lexeme.i == tag->lexeme.i )
         break;
      p = r;
   }
   assert(r != NULL);
   if( p == NULL )
      return;

   t = r->tree;
   r->tree = *node;
   r->next = (*node)->next;
   p->tree = t;
   (*node)->next = NULL;
   *node = r;

} /* RotateAttr() */

/**************************************************************** TagToBit */
static int TagToBit(Token *token)
{
   switch( token->type )
   {
      case TagB:        return 0x000400;
      case TagI:        return 0x000800;
      case TagS:        return 0x001000;
      case TagU:        return 0x002000;
      case TagEM:       return 0x004000;
      case TagTT:       return 0x008000;

      case TagColor:
         switch( token->lexeme.i )
         {
            case 'r':   return 0x010000;
            case 'g':   return 0x020000;
            case 'b':   return 0x040000;
            case 'c':   return 0x080000;
            case 'm':   return 0x100000;
            case 'y':   return 0x200000;
            case 'k':   return 0x400000;
            case 'w':   return 0x800000;
            default:
               break;
         }
         break;

      case TagSize:
         assert(token->lexeme.i >= '0' && token->lexeme.i <= '9');
         return 1 << (token->lexeme.i - '0');
      default:
         break;
   }
   return 0;

} /* TagToBit() */
